using System;
using System.Threading;

namespace Eratosthenes
{
  abstract class NumberStream {
    public abstract bool IsFinished();
    public abstract int Get();
  }

  class Sieve: NumberStream {
    NumberStream source;
    Thread t;
    int prime;
    bool hasPrime;
    int next;
    bool hasNext;
    bool finished;

    public Sieve(NumberStream prev) {
      source = prev;
      hasPrime = false; hasNext = false;
      finished = false;
      t = new Thread(new ThreadStart(Run));
      t.Start();
    }
    
    void Run() {
      if (!source.IsFinished()) {
        lock(this) { 
          prime = source.Get(); hasPrime = true; 
          Monitor.PulseAll(this); 
        }
      }
      while (!source.IsFinished()) {
        lock(this) {
          int i = source.Get();
          if (i % prime != 0) {
            while (hasNext) { Monitor.Wait(this); }
            next = i; hasNext = true;
            Monitor.PulseAll(this);
          }
        }
      }
      lock(this) { 
        finished = true; Monitor.PulseAll(this); 
      }
    }

    public bool HasPrime() {
      lock(this) {
        while (!hasPrime && !finished) { Monitor.Wait(this); }
      }
      return hasPrime;  
    }
    
    public int GetPrime() {
      return prime;
    }
    
    public override bool IsFinished() {
      lock(this) {
        while (!hasNext && !finished) { Monitor.Wait(this); }
        return !hasNext;
      }
    }

    public override int Get() {
      lock(this) {
        while (!hasNext) { Monitor.Wait(this); }
        int x = next; hasNext = false;
        Monitor.PulseAll(this);
        return x;
      }
    }
  }

  class NumberGenerator: NumberStream {
    const int N = 10000;
    Thread t;
    int next;
    bool hasNext;
    bool finished;

    public NumberGenerator() {
      hasNext = false;
      finished = false;
      t = new Thread(new ThreadStart(Run));
      t.Start();
    }

    void Run() {
      for (int i = 2; i <= N; i++) {
        lock(this) {
          while (hasNext) { Monitor.Wait(this); }
          next = i; hasNext = true;
          Monitor.PulseAll(this);
        }
      }
      lock(this) {
        finished = true;
        Monitor.PulseAll(this);
      }
    }

    public override bool IsFinished() {
      lock(this) {
        while (!hasNext && !finished) { Monitor.Wait(this); }
        return !hasNext;
      }
    }

    public override int Get() {
      lock(this) {
        while (!hasNext) { Monitor.Wait(this); }
        int x = next; hasNext = false;
        Monitor.PulseAll(this);
        return x;
      }
    }
  }

	class Eratosthenes
	{
		static void Main(string[] args)
		{
      DateTime startTime = DateTime.Now;
      NumberGenerator generator = new NumberGenerator();
      int i = 1;
      Sieve sieve = new Sieve(generator);
      while (sieve.HasPrime()) {
        int p = sieve.GetPrime();
        //Console.Write(p + " ");
        sieve = new Sieve(sieve); i++;
      }
      //Console.WriteLine();
      Console.WriteLine((DateTime.Now - startTime) + " ms");
    }
	}
}
