using System;
using System.Threading;

namespace ProducerConsumer
{
  class BoundedBuffer {
    const int C = 10;
    int[] a;
    int first, last;
    int nofProducers;
    
    public BoundedBuffer(int n) {
      a = new int[C];
      first = 0; last = 0; nofProducers = n;
    }

    public void Put(int x) {
      lock(this) {
        while (last-first >= C) { Monitor.Wait(this); }
        a[last % C] = x; last++;
        Monitor.PulseAll(this);
      }
    }

    public bool InputAvailable() {
      lock(this) {
        while ((first >= last) && (nofProducers > 0)) { Monitor.Wait(this); }
        return first < last;
      }
    }

    public void Get(out int x) {
      lock(this) {
        x = a[first % C]; first++; 
        Monitor.PulseAll(this);
      }
    }

    public void Finish() {
      lock(this) { 
        nofProducers--;
        if (nofProducers == 0) { Monitor.PulseAll(this); }
      }
    }
  }

  class Producer {
    BoundedBuffer buffer;
    Thread t;

    public Producer(BoundedBuffer b) {
      buffer = b;
      t = new Thread(new ThreadStart(Run));
    }

    public void Start() {
      t.Start();
    }
    
    public void Run() {
      for (int i = 1; i <= ProducerConsumerBenchmark.K; i++) {
        buffer.Put(i);
      }
      buffer.Finish();
    }

    public void Terminate() {
      t.Join();
    }
  }

  class Consumer {
    BoundedBuffer buffer;
    Thread t;

    public Consumer(BoundedBuffer b) {
      buffer = b;
      t = new Thread(new ThreadStart(Run));
    }

    public void Start() {
      t.Start();
    }
    
    public void Run() {
      while (buffer.InputAvailable()) {
        int x;
        buffer.Get(out x);
        // if (x% 100000 == 0) { Console.WriteLine(x); }
      }
    }

    public void Terminate() {
      t.Join();
    }
  }

	class ProducerConsumerBenchmark
	{
    public const int N = 5;
    public const int M = 5;
    public const int K = 1000000;

    static void Main(string[] args) {
      DateTime startTime = DateTime.Now;
      BoundedBuffer buffer = new BoundedBuffer(N);
      Producer[] producer = new Producer[N];
      Consumer[] consumer = new Consumer[M];
      for (int i = 0; i < N; i++) { producer[i] = new Producer(buffer); }
      for (int i = 0; i < M; i++) { consumer[i] = new Consumer(buffer); }
      for (int i = 0; i < N; i++) { producer[i].Start(); }
      for (int i = 0; i < M; i++) { consumer[i].Start(); }
      for (int i = 0; i < N; i++) { producer[i].Terminate(); }
      for (int i = 0; i < M; i++) { consumer[i].Terminate(); }
      Console.WriteLine((DateTime.Now - startTime) + " ms");
    }
  }
}
